/*
 * yasapi_player.c
 * Copyright (C) 2015-2016 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
// Rendering a Stream
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd316756%28v=vs.85%29.aspx

// Exclusive-Mode Streams
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd370844%28v=vs.85%29.aspx

#if ! defined (AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY) // {
#define AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY 0x08000000
#endif // }

#if ! defined(AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM) // {
#define AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM 0x80000000
#endif // }

///////////////////////////////////////////////////////////////////////////////
int PlayerCreate(Player *pPlayer, HINSTANCE hModule, const wchar_t *path,
    int debug)
{
  char dummy[MAX_PATH];
  int nLen;

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(pPlayer,sizeof *pPlayer);
  pPlayer->hDlgConfig=NULL;

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->options.path=path;
  OptionsCommonLoad(&pPlayer->options.common,L"    ",path,debug);
#if defined (YASAPI_DEBUG) // {
  pPlayer->options.common.nDebug=debug;
#endif // }
  DPUTS(debug,0,"  common options loaded\n");
  OptionsDeviceLoad(&pPlayer->options.device,L"    ",
      pPlayer->options.common.szId,path,debug);
  DPUTS(debug,0,"  device options loaded\n");

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->state=PLAYER_STATE_NULL;
  pPlayer->base.qVolume=1.0;

  /////////////////////////////////////////////////////////////////////////////
  nLen=GetModuleFileNameA(
    hModule,    // _In_opt_ HMODULE hModule,
    dummy,      // _Out_    LPTSTR  lpFilename,
    MAX_PATH-1  // _In_     DWORD   nSize
  );

  if (ERROR_INSUFFICIENT_BUFFER==GetLastError()) {
    DMESSAGEV(debug,"getting module name (%s)",YASAPI_MODULE);
    goto module;
  }

  nLen=strlen(dummy);

  if (NULL==(pPlayer->base.pszModuleName=malloc(nLen+1))) {
    DMESSAGE(debug,"allocating buffer for module name");
    goto malloc;
  }

  strcpy(pPlayer->base.pszModuleName,dummy);
  pPlayer->base.pszFileName=pPlayer->base.pszModuleName;
  pPlayer->base.pszFileName+=strlen(pPlayer->base.pszFileName);

  while (pPlayer->base.pszModuleName<pPlayer->base.pszFileName
      &&'/'!=pPlayer->base.pszFileName[-1]
      &&'\\'!=pPlayer->base.pszFileName[-1]
      &&':'!=pPlayer->base.pszFileName[-1]) {
    --pPlayer->base.pszFileName;
  }

  DPRINTF(debug,0,"  got module name: \"%s\"\n",pPlayer->base.pszFileName);
  pPlayer->base.hModule=hModule;

  /////////////////////////////////////////////////////////////////////////////
  if (QueueCreate(&pPlayer->base.queue,debug)<0) {
    DMESSAGE(debug,"creating queue");
    goto queue;
  }

  DPUTS(debug,0,"  queue created\n");

  /////////////////////////////////////////////////////////////////////////////
  if (StoreCreate(&pPlayer->base.store,debug)<0) {
    DMESSAGE(debug,"creating store");
    goto store;
  }

  DPUTS(debug,0,"  store created\n");

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->base.hEvent=CreateEvent(
    0,        // _In_opt_ LPSECURITY_ATTRIBUTES lpEventAttributes,
    FALSE,    // _In_     BOOL                  bManualReset,
    FALSE,    // _In_     BOOL                  bInitialState,
    NULL      // _In_opt_ LPCTSTR               lpName
  );

  if (NULL==pPlayer->base.hEvent) {
    DMESSAGE(debug,"creating event");
    goto event;
  }
  
  DPUTS(debug,0,"  event created\n");

  // create the timer /////////////////////////////////////////////////////////
  pPlayer->base.hTimer=CreateWaitableTimer(
    NULL,       // _In_opt_  LPSECURITY_ATTRIBUTES lpTimerAttributes,
    FALSE,      // _In_      BOOL bManualReset,
    NULL        // _In_opt_  LPCTSTR lpTimerName
  );

  if (NULL==pPlayer->base.hTimer) {
    DMESSAGE(debug,"creating timer");
    goto timer;
  }

  DPUTS(debug,0,"  waitable timer created\n");

  // create the thread ////////////////////////////////////////////////////////
  pPlayer->base.hThread=CreateThread(
    NULL,           // _In_opt_   LPSECURITY_ATTRIBUTES lpThreadAttributes,
    0,              // _In_       SIZE_T dwStackSize,
    PlayerThread,   // _In_       LPTHREAD_START_ROUTINE lpStartAddress,
    pPlayer,        // _In_opt_   LPVOID lpParameter,
    0,              // _In_       DWORD dwCreationFlags,
    NULL            // _Out_opt_  LPDWORD lpThreadId
  );

  if (NULL==pPlayer->base.hThread) {
    DMESSAGE(debug,"creating thread");
    goto thread;
  }

  DPUTS(debug,0,"  thread created\n");

  /////////////////////////////////////////////////////////////////////////////
  ++pPlayer->state;

  /////////////////////////////////////////////////////////////////////////////
//DPRINTF(debug,0,"  %s 1\n",__func__);
  if (PlayerDispatchSync(pPlayer,0,PlayerRun)<0) {
    DMESSAGE(debug,"running");
    goto run;
  }
//DPRINTF(debug,0,"  %s 2\n",__func__);

  DPUTS(debug,0,"  run request send\n");

  return 0;
// cleanup:
  //PlayerDispatchSync(pPlayer,0,PlayerStop);
run:
  CloseHandle(pPlayer->base.hThread);
thread:
  CloseHandle(pPlayer->base.hTimer);
timer:
  CloseHandle(pPlayer->base.hEvent);
event:
  StoreDestroy(&pPlayer->base.store,debug);
store:
  QueueDestroy(&pPlayer->base.queue,debug);
queue:
  free(pPlayer->base.pszModuleName);
malloc:
module:
  return 0;
}

void PlayerDestroy(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#else // } {
  int debug=0;
#endif // }

  if (PLAYER_STATE_RUN<pPlayer->state)
    PlayerDispatchSync(pPlayer,0,PlayerClose);

  if (PLAYER_STATE_NULL<pPlayer->state) {
    ///////////////////////////////////////////////////////////////////////////
    PlayerDispatchSync(pPlayer,1,PlayerStop);
    WaitForSingleObject(pPlayer->base.hThread,INFINITE);

    ///////////////////////////////////////////////////////////////////////////
    --pPlayer->state;

    ///////////////////////////////////////////////////////////////////////////
    DPUTS(debug,0,"  destroying thread\n");
    CloseHandle(pPlayer->base.hThread);

    DPUTS(debug,0,"  destroying timer\n");
    CloseHandle(pPlayer->base.hTimer);

    DPUTS(debug,0,"  destroying event\n");
    CloseHandle(pPlayer->base.hEvent);

    DPUTS(debug,0,"  destroying store\n");
    StoreDestroy(&pPlayer->base.store,debug);

    DPUTS(debug,0,"  destroying queue\n");
    QueueDestroy(&pPlayer->base.queue,debug);

    DPUTS(debug,0,"  freeing module name\n");
    free(pPlayer->base.pszModuleName);
  }
}

///////////////////////////////////////////////////////////////////////////////
int PlayerRun(Player *pPlayer, Request *pRequest)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  IMMDeviceEnumerator *pEnumerator;
  HRESULT hr;

  DPRINTF(debug,0,"  > %s <\n",__func__);

  /////////////////////////////////////////////////////////////////////////////
  if (PLAYER_STATE_BASE!=pPlayer->state) {
    DMESSAGEV(debug,"unexpected state: %d",pPlayer->state);
    goto state;
  }

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&pPlayer->run,sizeof pPlayer->run);

  // create a device enumerator ///////////////////////////////////////////////
  hr=CoCreateInstance(
    &CLSID_MMDeviceEnumerator,  // _In_   REFCLSID rclsid,
    NULL,                       // _In_   LPUNKNOWN pUnkOuter,
    CLSCTX_ALL,                 // _In_   DWORD dwClsContext,
    &IID_IMMDeviceEnumerator,   // _In_   REFIID riid,
    (void**)&pEnumerator        // _Out_  LPVOID *ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,REGDB_E_CLASSNOTREG,hr);
    DERROR(debug,CLASS_E_NOAGGREGATION,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"creating device enumerator");
    goto enumerator;
  }

  pPlayer->run.pEnumerator=pEnumerator;
  DPUTS(debug,0,"  device enumerator created\n");

#if defined (YASAPI_NOTIFY) // {
  /////////////////////////////////////////////////////////////////////////////
  if (PlayerAddNotify(pPlayer)<0)
    goto notify;
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  ++pPlayer->state;

  return 0;
//cleanup:
#if defined (YASAPI_NOTIFY) // {
  PlayerRemoveNotify(pPlayer);
notify:
#endif // }
  pEnumerator->lpVtbl->Release(pEnumerator);
enumerator:
state:
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
int PlayerStop(Player *pPlayer, Request *pRequest)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  IMMDeviceEnumerator *pEnumerator=pPlayer->run.pEnumerator;

  DPRINTF(debug,0,"  > %s <\n",__func__);

  if (PLAYER_STATE_BASE<pPlayer->state) {
    ///////////////////////////////////////////////////////////////////////////
    --pPlayer->state;

#if defined (YASAPI_NOTIFY) // {
    ///////////////////////////////////////////////////////////////////////////
    DPUTS(debug,0,"  removing notify\n");
    PlayerRemoveNotify(pPlayer);
#endif // }

    DPUTS(debug,0,"  destroying device enumerator\n");
    pEnumerator->lpVtbl->Release(pEnumerator);
  }

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
static UINT64 PlayerGetPosition(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  IAudioClock *pClock=pPlayer->open.pClock;
  UINT64 u64Position;
  HRESULT hr;

  // get position /////////////////////////////////////////////////////////////
  hr=pClock->lpVtbl->GetPosition(pClock,
    &u64Position,         // [out] UINT64 *pu64Position,
    NULL                  // [out] UINT64 *pu64QPCPosition
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"getting position");
    goto position;
  }

  return u64Position;
position:
  return -1;
}

static UINT64 PlayerGetFrequency(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  IAudioClock *pClock=pPlayer->open.pClock;
  UINT64 u64Frequency;
  HRESULT hr;

  // get frequency ////////////////////////////////////////////////////////////
  hr=pClock->lpVtbl->GetFrequency(pClock,
    &u64Frequency         // [out] UINT64 *pu64Frequency
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"getting frequency");
    goto frequency;
  }

  return u64Frequency;
frequency:
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_GAPLESS) // {
static int PlayerResetGaplessOffset(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }

  pPlayer->open.nGaplessOffset=pPlayer->options.device.nGaplessOffset;

  switch (pPlayer->open.nGaplessOffset) {
  case YASAPI_GAPLESS_OFFSET_POSITION:
    pPlayer->open.u64Position=0ll;

    if ((pPlayer->open.u64Frequency=PlayerGetFrequency(pPlayer))<=0) {
      DMESSAGE(debug,"getting frequency\n");
      goto frequency;
    }

    if (pPlayer->options.common.bGapless)
      DPRINTF(debug,0,"  gapless offset reset to position\n");

    break;
  case YASAPI_GAPLESS_OFFSET_TIME:
    pPlayer->open.qTime=0.0;

    if (pPlayer->options.common.bGapless)
      DPRINTF(debug,0,"  gapless offset reset to time\n");

    break;
  default:
    DMESSAGEV(debug,"tag mismatch: %d\n",pPlayer->open.nGaplessOffset);
    goto tag;
  }

  return 0;
tag:
frequency:
position:
  return -1;
}
#endif // }

int PlayerOpen(Player *pPlayer, Request *pRequest)
{
  const wchar_t *path=pPlayer->options.path;
#if defined (YASAPI_DEBUG) // {
  int *pDebug=&pPlayer->options.common.nDebug;
  int nDebug=*pDebug;
#else // } {
  int *pDebug=NULL;
  int nDebug=0;
#endif // }
  PlayerDevice *pPlayerDevice=va_arg(pRequest->ap,PlayerDevice *);
  int samplerate=va_arg(pRequest->ap,int);
  int numchannels=va_arg(pRequest->ap,int);
  int bitspersamp=va_arg(pRequest->ap,int);
#if defined (YASAPI_NOTIFY) // {
  Ring *pRing=va_arg(pRequest->ap,Ring *);
#endif // }
  IMMDeviceEnumerator *pEnumerator=pPlayer->run.pEnumerator;
  WAVEFORMATEXTENSIBLE *pwfxx=&pPlayer->open.wfxx;
  WAVEFORMATEX *pwfx=&pwfxx->Format;
#if defined (YASAPI_SORROUND) // {
  int bSorround;
#endif // }
  WAVEFORMATEX *pClosestMatch;
  IMMDevice *pDevice;
  IAudioClient *pClient;
  IAudioRenderClient *pRender;
  IAudioClock *pClock;
  const Strategy *pStrategy;
  AUDCLNT_SHAREMODE eShareMode;
  REFERENCE_TIME hnsDefaultDevicePeriod;
  REFERENCE_TIME hnsMinimumDevicePeriod;
  REFERENCE_TIME hnsDevicePeriod;
  DWORD dwStreamFlags;
  double qShareSize;
  double qRingFill;
  double qRingSize;
  UINT32 uBufSize;
  int retry;
  HRESULT hr;

  DPRINTF(nDebug,0,"  > %s <\n",__func__);

  /////////////////////////////////////////////////////////////////////////////
  if (PLAYER_STATE_RUN!=pPlayer->state) {
    DMESSAGEV(nDebug,"unexpected state: %d (%s)",pPlayer->state,__func__);
    goto state;
  }

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&pPlayer->open,sizeof pPlayer->open);

  pPlayer->device=*pPlayerDevice;
  ZeroMemory(pPlayerDevice,sizeof *pPlayerDevice);
  pPlayerDevice=&pPlayer->device;

  if (!pPlayerDevice->pDevice&&PlayerDeviceGet(pPlayerDevice,pPlayer)<0) {
    DMESSAGE(nDebug,"  getting the device");
    goto device;
  }

  DPUTS(nDebug,0,"  got the device\n");
  pDevice=pPlayerDevice->pDevice;

  OptionsDeviceLoad(&pPlayer->options.device,L"    ",
      pPlayerDevice->szId,path,nDebug);
  DPUTS(nDebug,0,"  device options loaded\n");

  /////////////////////////////////////////////////////////////////////////////
  qShareSize=pPlayer->options.device.qShareSize;
  qRingFill=pPlayer->options.device.ring.qFill;
  qRingSize=pPlayer->options.device.ring.qSize;

  /////////////////////////////////////////////////////////////////////////////
  DPRINTF(nDebug,0,"  samplerate: %d\n",samplerate);

  if (1==numchannels&&pPlayer->options.common.bMono2Stereo) {
    DPUTS(nDebug,0,"  numchannels: 1 -> 2\n");
    pPlayer->open.nShift=1;
    numchannels=2;
  }
  else {
    DPRINTF(nDebug,0,"  numchannels: %d\n",numchannels);
    pPlayer->open.nShift=0;
  }

  DPRINTF(nDebug,0,"  bitspersamp: %d\n",bitspersamp);
#if defined (YASAPI_SORROUND) // {
  bSorround=pPlayer->open.bSorround=pPlayer->options.common.bSorround;
  WFXXSetup(pwfxx,samplerate,numchannels,bitspersamp,bSorround,FALSE);
#else // } {
  WFXXSetup(pwfxx,samplerate,numchannels,bitspersamp,FALSE,FALSE);
#endif // }
  pPlayer->open.wBytesPerSample=pwfx->wBitsPerSample>>3;

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->open.pStrategy=pStrategy
      =pPlayer->options.device.bPull?&gcStrategyPull:&gcStrategyPush;

  // setup share mode /////////////////////////////////////////////////////////
  switch (pPlayer->options.device.eShareMode) {
  case YASAPI_SHAREMODE_EXCLUSIVE:
    DPUTS(nDebug,0,"  YASAPI_SHAREMODE_EXCLUSIVE\n");
    eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
    break;
  case YASAPI_SHAREMODE_AUTOMATIC:
    DPUTS(nDebug,0,"  YASAPI_SHAREMODE_AUTOMATIC\n");
    eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
    break;
  default:
    DPUTS(nDebug,0,"  AUDCLNT_SHAREMODE_SHARED\n");
    eShareMode=AUDCLNT_SHAREMODE_SHARED;
    break;
  }

  /////////////////////////////////////////////////////////////////////////////
  hnsDevicePeriod=0;
  retry=0;
retry:
  // remember the share mode //////////////////////////////////////////////////
  pPlayer->open.eShareMode=eShareMode;

  // activate the audio client ////////////////////////////////////////////////
  hr=pDevice->lpVtbl->Activate(pDevice,
    &IID_IAudioClient,    // [in]   REFIID iid,
    CLSCTX_ALL,           // [in]   DWORD dwClsCtx,
    NULL,                 // [in]   PROPVARIANT *pActivationParams,
    (void **)&pClient     // [out]  void **ppInterface
  );

  if (FAILED(hr)) {
    DERROR(nDebug,E_NOINTERFACE,hr);
    DERROR(nDebug,E_POINTER,hr);
    DERROR(nDebug,E_INVALIDARG,hr);
    DERROR(nDebug,E_OUTOFMEMORY,hr);
    DERROR(nDebug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DUNKNOWN(nDebug,hr);
    DMESSAGE(nDebug,"activating audio client");
    goto client;
  }

  pPlayer->open.pClient=pClient;
  DPUTS(nDebug,0,"  audio client created\n");

  // format supported? ////////////////////////////////////////////////////////
  if (pPlayer->options.common.bFormatSupported) {
    pClosestMatch=NULL;

    hr=pClient->lpVtbl->IsFormatSupported(pClient,
      eShareMode,         // [in]        AUDCLNT_SHAREMODE ShareMode,
      pwfx,               // [in]  const WAVEFORMATEX      *pFormat,
      AUDCLNT_SHAREMODE_EXCLUSIVE==eShareMode?NULL:&pClosestMatch
                          // [out]       WAVEFORMATEX      **ppClosestMatch
    );

    if (pClosestMatch) {
      DPRINTF(nDebug,0,"  closest match: %d, %d\n",
          pClosestMatch->wBitsPerSample,
          pClosestMatch->nSamplesPerSec);
      CoTaskMemFree(pClosestMatch);
    }

    switch (hr) {
    case S_OK:
      break;
    case S_FALSE:
      DERROR(nDebug,S_FALSE,hr);
      DMESSAGE(nDebug,"format not supported");
      goto support;
    default:
      if (FAILED(hr)) {
        DERROR(nDebug,AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
        DERROR(nDebug,E_POINTER,hr);
        DERROR(nDebug,E_INVALIDARG,hr);
        DERROR(nDebug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
        DERROR(nDebug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
        DUNKNOWN(nDebug,hr);
        DMESSAGE(nDebug,"format not supported");
        goto support;
      }

      break;
    }

    DPUTS(nDebug,0,"  format supported\n");
  }

  // setup the stream flags ///////////////////////////////////////////////////
  dwStreamFlags=0;

  if (AUDCLNT_SHAREMODE_SHARED==eShareMode
        &&pPlayer->options.device.bAutoConvertPCM) {
      dwStreamFlags|=AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM;

     if (pPlayer->options.device.bSRCDefaultQuality)
       dwStreamFlags|=AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;
  }

  // if needed, get the minimum device period (low latency) ///////////////////
  // we may have it already because of not aligned buffer error, cf. below.
  if (0==hnsDevicePeriod&&pStrategy->NeedDevicePeriod(pPlayer)) {
    hr=pClient->lpVtbl->GetDevicePeriod(pClient,
      &hnsDefaultDevicePeriod,
                          // [out]  REFERENCE_TIME *phnsDefaultDevicePeriod,
      &hnsMinimumDevicePeriod
                          // [out]  REFERENCE_TIME *phnsMinimumDevicePeriod
    );

    if (FAILED(hr)) {
      DERROR(nDebug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
      DERROR(nDebug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
      DERROR(nDebug,E_POINTER,hr);
      DUNKNOWN(nDebug,hr);
      DMESSAGE(nDebug,"getting device period");
      goto period;
    }

    switch (pPlayer->options.device.eDevicePeriod) {
    case YASAPI_DEVICE_PERIOD_DEFAULT:
      hnsDevicePeriod=hnsDefaultDevicePeriod;
      break;
    case YASAPI_DEVICE_PERIOD_MINIMUM:
      hnsDevicePeriod=hnsMinimumDevicePeriod;
      break;
    default:
      hnsDevicePeriod=hnsMinimumDevicePeriod;
      break;
    }

    DPUTS(nDebug,0,"  got device period\n");
    DPRINTF(nDebug,0,"    default: %I64d hns\n",hnsDefaultDevicePeriod);
    DPRINTF(nDebug,0,"    minimum: %I64d hns\n",hnsMinimumDevicePeriod);
    DPRINTF(nDebug,0,"    choosen: %I64d hns\n",hnsDevicePeriod);

    if (1.0<qShareSize) {
      hnsDevicePeriod=qShareSize*hnsDevicePeriod+0.5;
      DPRINTF(nDebug,0,"  device period scaled by %.02f"
          " (device period: %I64d)\n",
          qShareSize,hnsDevicePeriod);
    }
  }

  // initialize the audio client //////////////////////////////////////////////
  hr=pPlayer->open.pStrategy->Initialize(pPlayer,dwStreamFlags,
      hnsDevicePeriod);

  if (FAILED(hr)) {
    if (AUDCLNT_SHAREMODE_EXCLUSIVE==eShareMode) {
      switch (hr) {
      case AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED:
        if (0<retry)
          break;

        DPUTS(nDebug,0,"  failed with AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED\n");
        DPUTS(nDebug,0,"  trying to get an aligned buffer size\n");
        retry=1;

        // 1. Call IAudioClient::GetBufferSize and receive the
        //    next-highest-aligned buffer size (in frames).
        hr=pClient->lpVtbl->GetBufferSize(pClient,
          &pPlayer->open.uFrames    // [out] UINT32 *pNumBufferFrames
        );

        if (FAILED(hr)) {
          DERROR(nDebug,AUDCLNT_E_NOT_INITIALIZED,hr);
          DERROR(nDebug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
          DERROR(nDebug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
          DERROR(nDebug,E_POINTER,hr);
          DUNKNOWN(nDebug,hr);
          DMESSAGE(nDebug,"getting aligned number of frames");
          goto init;
        }

        DPRINTF(nDebug,0,"  got aligned number of frames: %u\n",
            pPlayer->open.uFrames);

        // 3. Calculate the aligned buffer size in 100-nansecond units (hns).
        //    The buffer size is (REFERENCE_TIME)((10000.0 * 1000 /
        //    WAVEFORMATEX.nSamplesPerSecond * nFrames) + 0.5). In this
        //    formula, nFrames is the buffer size retrieved by GetBufferSize.
        hnsDevicePeriod=YASAPI_FRAMES_TIMER(pPlayer->open.uFrames,pwfx)+0.5;

#if 0 // {
        // save hnsDevicePeriod into device options.
        OptionsSetDevicePeriod(pPlayer->options.hnsDevicePeriod);
#endif // }

        // 2. Call IAudioClient::Release to release the audio
        //    client used in the previous call that returned
        //    AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED.
        DPUTS(nDebug,0,"  destroying audio client\n");
        pClient->lpVtbl->Release(pClient);

        // 4. Call the IMMDevice::Activate method with parameter iid set
        //    to REFIID IID_IAudioClient to create a new audio client.
        // 5. Call Initialize again on the created audio client and
        //    specify the new buffer size and periodicity.
        DPUTS(nDebug,0,"  retrying with aligned buffer\n");
        goto retry;
      default:
        if (YASAPI_SHAREMODE_AUTOMATIC!=pPlayer->options.device.eShareMode)
          break;

        DPUTS(nDebug,0,"  initializing audio client failed \n");
        DPUTS(nDebug,0,"  destroying audio client\n");
        pClient->lpVtbl->Release(pClient);

        DPUTS(nDebug,0,"  retrying in shared mode\n");
        eShareMode=AUDCLNT_SHAREMODE_SHARED;
        goto retry;
      }
    }

    DERROR(nDebug,AUDCLNT_E_ALREADY_INITIALIZED,hr);
    DERROR(nDebug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(nDebug,AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED,hr);
    DERROR(nDebug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(nDebug,AUDCLNT_E_CPUUSAGE_EXCEEDED,hr);
    DERROR(nDebug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(nDebug,AUDCLNT_E_DEVICE_IN_USE,hr);
    DERROR(nDebug,AUDCLNT_E_ENDPOINT_CREATE_FAILED,hr);
    DERROR(nDebug,AUDCLNT_E_INVALID_DEVICE_PERIOD,hr);
    DERROR(nDebug,AUDCLNT_E_UNSUPPORTED_FORMAT,hr);
    DERROR(nDebug,AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED,hr);
    DERROR(nDebug,AUDCLNT_E_BUFDURATION_PERIOD_NOT_EQUAL,hr);
    DERROR(nDebug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(nDebug,E_POINTER,hr);
    DERROR(nDebug,E_INVALIDARG,hr);
    DERROR(nDebug,E_OUTOFMEMORY,hr);
    DUNKNOWN(nDebug,hr);
    DMESSAGE(nDebug,"initializing audio client");
    goto init;
  }

  DPUTS(nDebug,0,"  audio client initialized\n");

  // set the event handle /////////////////////////////////////////////////////
  hr=pStrategy->SetEvent(pPlayer);

  if (FAILED(hr)) {
    DERROR(nDebug,E_INVALIDARG,hr);
    DERROR(nDebug,AUDCLNT_E_EVENTHANDLE_NOT_EXPECTED,hr);
    DERROR(nDebug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(nDebug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(nDebug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(nDebug,hr);
    DMESSAGE(nDebug,"setting event handle");
    goto event;
  }

  // get the buffer size //////////////////////////////////////////////////////
  //
  // This method retrieves the length of the endpoint buffer shared
  // between the client application and the audio engine. The length
  // is expressed as the number of audio frames the buffer can hold.
  // The size in bytes of an audio frame is calculated as the number
  // of channels in the stream multiplied by the sample size per channel.
  // For example, the frame size is four bytes for a stereo (2-channel)
  // stream with 16-bit samples.
  hr=pClient->lpVtbl->GetBufferSize(pClient,
    &pPlayer->open.uFrames    // [out]  UINT32 *pNumBufferFrames
  );

  if (FAILED(hr)) {
    DERROR(nDebug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(nDebug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(nDebug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(nDebug,E_POINTER,hr);
    DUNKNOWN(nDebug,hr);
    DMESSAGE(nDebug,"getting buffer size");
    goto size;
  }

  DPRINTF(nDebug,0,"  SHARED BUFFER SIZE: %d frames\n",pPlayer->open.uFrames);
  pPlayer->open.time=YASAPI_FRAMES_TIMER(pPlayer->open.uFrames,pwfx);

  // get the render client ////////////////////////////////////////////////////
  hr=pClient->lpVtbl->GetService(pClient,
    &IID_IAudioRenderClient,  // [in]   REFIID riid,
    (void**)&pRender          // [out]  void **ppv
  );

  if (FAILED(hr)) {
    DERROR(nDebug,E_POINTER,hr);
    DERROR(nDebug,E_NOINTERFACE,hr);
    DERROR(nDebug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(nDebug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(nDebug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(nDebug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(nDebug,hr);
    DMESSAGE(nDebug,"getting audio render client");
    goto render;
  }

  pPlayer->open.pRender=pRender;
  DPUTS(nDebug,0,"  audio render client created\n");

  // get the audio clock //////////////////////////////////////////////////////
  hr=pClient->lpVtbl->GetService(pClient,
    &IID_IAudioClock,         // [in]   REFIID riid,
    (void**)&pClock           // [out]  void **ppv
  );

  if (FAILED(hr)) {
    DERROR(nDebug,E_POINTER,hr);
    DERROR(nDebug,E_NOINTERFACE,hr);
    DERROR(nDebug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(nDebug,AUDCLNT_E_WRONG_ENDPOINT_TYPE,hr);
    DERROR(nDebug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(nDebug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(nDebug,hr);
    DMESSAGE(nDebug,"getting audio clock");
    goto clock;
  }

  pPlayer->open.pClock=pClock;
  DPUTS(nDebug,0,"  audio clock created\n");

  // reset the client /////////////////////////////////////////////////////////
  hr=pClient->lpVtbl->Reset(pClient);

  if (FAILED(hr)) {
    DERROR(nDebug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(nDebug,AUDCLNT_E_NOT_STOPPED,hr);
    DERROR(nDebug,AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
    DERROR(nDebug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(nDebug,hr);
    DMESSAGE(nDebug,"re-setting audio client");
    goto reset;
  }

  DPUTS(nDebug,0,"  audio client reset\n");

  // setup the ring buffer ////////////////////////////////////////////////////
  pPlayer->open.uFramesMin=qRingFill*pPlayer->open.uFrames+0.5;
  DPRINTF(nDebug,0,"  START AUDIO CLIENT: %d frames\n",
      pPlayer->open.uFramesMin);

  uBufSize=qRingSize*pPlayer->open.uFrames+0.5;
  // "nBlockAlign" reflects "bMono2Stereo".
  uBufSize*=pwfx->nBlockAlign;

#if defined (YASAPI_NOTIFY)
  if (pRing) {
    pPlayer->open.ring=*pRing;

    if (RingReallocSize(&pPlayer->open.ring,uBufSize,pDebug)<0) {
      DMESSAGE(nDebug,"re-allocating ring buffer\n");
      goto ring;
    }
  }
  else {
#endif
  if (RingCreate(&pPlayer->open.ring,uBufSize,pDebug)<0) {
    DMESSAGE(nDebug,"creating ring buffer\n");
    goto ring;
  }
#if defined (YASAPI_NOTIFY)
  }
#endif

  pPlayer->open.ring.Copy=PlayerCopyMemory;
  pPlayer->open.ring.pClient=pPlayer;
#if defined (YASAPI_RING_REALLOC) // {
  pPlayer->open.dwBufSize=uBufSize;
#endif // }
  DPRINTF(nDebug,0,"  ring buffer created (%d bytes)\n",uBufSize);

#if defined (YASAPI_GAPLESS) // {
  if (PlayerResetGaplessOffset(pPlayer)<0) {
    DMESSAGE(nDebug,"re-setting gapless offset\n");
    goto offset;
  }
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  ++pPlayer->state;

  return YASAPI_FRAMES_MS(pPlayer->open.uFramesMin,pwfx);
#if defined (YASAPI_GAPLESS) // {
offset:
  RingDestroy(&pPlayer->open.ring);
#endif // }
ring:
reset:
  pClock->lpVtbl->Release(pClock);
clock:
  pRender->lpVtbl->Release(pRender);
render:
size:
event:
init:
period:
support:
  pClient->lpVtbl->Release(pClient);
client:
id:
  PlayerDeviceDestroy(pPlayerDevice,pPlayer);
device:
state:
  return -1;
}

static void PlayerKill(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  const Strategy *pStrategy=pPlayer->open.pStrategy;
  IAudioClient *pClient=pPlayer->open.pClient;

#if defined (YASAPI_GAPLESS) // {
retry:
#endif // }
  switch (pPlayer->state) {
#if defined (YASAPI_GAPLESS) // {
  case PLAYER_STATE_RESET:
    DPRINTF(debug,0,"  %s: %d -> %d\n",__func__,
        pPlayer->state_save,pPlayer->state);
    pPlayer->state=pPlayer->state_save;
    pPlayer->state_save=0;
    goto retry;
#endif // }
  case PLAYER_STATE_EOT:
    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_DRAIN_3:
    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_DRAIN_2:
    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_DRAIN_1:
    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_PLAY:
    --pPlayer->state;
    DPRINTF(debug,0,"  %s: cancelling timer\n",__func__);
    CancelWaitableTimer(pPlayer->base.hTimer);
    DPRINTF(debug,0,"  %s: stopping audio client\n",__func__);
    pClient->lpVtbl->Stop(pClient);
    PlayerPostUpdate(pPlayer,0,0);
    // intentional fall throuh, no break!
  case PLAYER_STATE_UNDERFLOW:
    --pPlayer->state;
    // intentional fall throuh, no break!
  case PLAYER_STATE_PAUSE:
    --pPlayer->state;
    // intentional fall throuh, no break!
  default:
    break;
  }
}

#if defined (YASAPI_GAPLESS) // {
int PlayerReset(Player *pPlayer, Request *pRequest)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  IAudioClient *pClient=pPlayer->open.pClient;
  WAVEFORMATEXTENSIBLE *pwfxx=&pPlayer->open.wfxx;
  WAVEFORMATEX *pwfx=&pwfxx->Format;
  UINT32 u32FramesPadding;
  UINT32 u32FramesWritten;
  UINT32 u32FramesWaiting;
  UINT64 u64PositionWaiting;
  UINT64 u64Frequency;
  UINT64 u64Position;

  DPRINTF(debug,0,"  > %s <\n",__func__);
  DPUTS(debug,0,"  re-using previous device\n");

  if (PlayerResetGaplessOffset(pPlayer)<0) {
    DMESSAGE(debug,"re-setting gapless offset\n");
    goto offset;
  }

  if ((u32FramesPadding=PlayerGetFramesPadding(pPlayer))<0) {
    DMESSAGE(debug,"getting frames padding");
    goto padding;
  }

  u32FramesWritten=pPlayer->open.ring.dwWritten/pwfx->nBlockAlign;
  u32FramesWaiting=u32FramesPadding+u32FramesWritten;

  switch (pPlayer->open.nGaplessOffset) {
  case YASAPI_GAPLESS_OFFSET_POSITION:
    u64Frequency=pPlayer->open.u64Frequency;
    break;
  case YASAPI_GAPLESS_OFFSET_TIME:
    if ((u64Frequency=PlayerGetFrequency(pPlayer))<0) {
      DMESSAGE(debug,"getting frequency");
      goto frequency;
    }

    break;
  default:
    goto offset;
  }

  u64PositionWaiting=MulDiv(u32FramesWaiting*pwfx->nChannels,
      u64Frequency,pwfx->nAvgBytesPerSec);

  if ((u64Position=PlayerGetPosition(pPlayer))<0) {
    DMESSAGE(debug,"getting position");
    goto position;
  }

  switch (pPlayer->open.nGaplessOffset) {
  case YASAPI_GAPLESS_OFFSET_POSITION:
    pPlayer->open.u64Position=u64Position+u64PositionWaiting;
    DPRINTF(debug,0,"  position offset: %I64d\n",pPlayer->open.u64Position);
    break;
  case YASAPI_GAPLESS_OFFSET_TIME:
    pPlayer->open.qTime=(double)(u64Position+u64PositionWaiting)/u64Frequency;
    DPRINTF(debug,0,"  time offset: %f\n",pPlayer->open.qTime);
    break;
  default:
    goto offset;
  }

  return YASAPI_FRAMES_MS(pPlayer->open.uFramesMin,pwfx);
position:
frequency:
padding:
offset:
  return -1;
}
#endif // }

int PlayerClose(Player *pPlayer, Request *pRequest)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
#if defined (YASAPI_NOTIFY) // {
  const IAudioClient *pRefClient=va_arg(pRequest->ap,const IAudioClient *);
#endif // }
  IAudioClock *pClock=pPlayer->open.pClock;
  IAudioRenderClient *pRender=pPlayer->open.pRender;
  IAudioClient *pClient=pPlayer->open.pClient;
  PlayerDevice *pPlayerDevice=&pPlayer->device;

  DPRINTF(debug,0,"  > %s <\n",__func__);

  if (PLAYER_STATE_RUN<pPlayer->state) {
    ///////////////////////////////////////////////////////////////////////////
    PlayerKill(pPlayer);

    ///////////////////////////////////////////////////////////////////////////
    --pPlayer->state;

    DPUTS(debug,0,"  destroying ring buffer\n");
    RingDestroy(&pPlayer->open.ring);

    DPUTS(debug,0,"  destroying audio clock\n");
    pClock->lpVtbl->Release(pClock);

    DPUTS(debug,0,"  destroying render client\n");
    pRender->lpVtbl->Release(pRender);

#if defined (YASAPI_NOTIFY)
    if (pRefClient!=pClient) {
#endif
      DPUTS(debug,0,"  destroying audio client\n");
      pClient->lpVtbl->Release(pClient);
      DPUTS(debug,0,"  done\n");
#if defined (YASAPI_NOTIFY)
    }
#endif

    PlayerDeviceDestroy(pPlayerDevice,pPlayer);
  }

  return 0;
}

int PlayerCanWrite(Player *pPlayer, Request *pRequest)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
  WAVEFORMATEX *pwfx=&pPlayer->open.wfxx.Format;
#endif // }
  DWORD dwAvailable;

  if (pPlayer->state<PLAYER_STATE_OPEN) {
#if ! defined (YASAPI_NOTIFY) // {
    DPRINTF(debug,0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
#endif // }
    dwAvailable=0;
  }
  else {
#if defined (YASAPI_RING_REALLOC) // {
    if (pPlayer->open.dwBufSize<pPlayer->open.ring.dwWritten)
      dwAvailable=0;
    else
      dwAvailable=pPlayer->open.dwBufSize-pPlayer->open.ring.dwWritten;
#else // } {
    dwAvailable=pPlayer->open.ring.dwAvailable;
#endif // }
    dwAvailable>>=pPlayer->open.nShift;
  }

  return dwAvailable;
}

//#define YASAPI_AUDCLNT_E_BUFFER_SIZE_ERROR
static int PlayerPlay(Player *pPlayer, UINT32 uFramesPadding)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  WAVEFORMATEX *pwfx=&pPlayer->open.wfxx.Format;
  int bNeedPadding=pPlayer->open.pStrategy->NeedPadding(pPlayer);
  IAudioRenderClient *pRender=pPlayer->open.pRender;
  Ring *pRing=&pPlayer->open.ring;
#if defined (YASAPI_AUDCLNT_E_BUFFER_SIZE_ERROR) // {
  UINT32 uFramesWrite=pPlayer->open.uFrames-uFramesPadding;
  DWORD dwSize=uFramesWrite*pwfx->nBlockAlign;
  DWORD dwWrite=pPlayer->open.ring.dwWritten;
#else // } {
  UINT32 uFramesWrite;
  DWORD dwSize;
  DWORD dwWrite;
#endif // }
  DWORD dwWriteSize;
  DWORD dwFlags=0;
#if defined (YASAPI_RING_DWORD) // {
  DWORD dwFramesRead;
#else // } {
  UINT32 uFramesRead;
#endif // }
  BYTE *pData;
  WORD wRing,wShared;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  if (pPlayer->state<PLAYER_STATE_PLAY) {
    DPRINTF(debug,0,"Warning: unexpected state: %d: \"%s\" (%d).\n",
        pPlayer->state,basename(__FILE__),__LINE__);
    PlayerKill(pPlayer);
    goto state;
  }

#if ! defined (YASAPI_AUDCLNT_E_BUFFER_SIZE_ERROR) // {
  uFramesWrite=pPlayer->open.uFrames;

  if (!bNeedPadding)
  	uFramesWrite-=uFramesPadding;

  dwSize=uFramesWrite*pwfx->nBlockAlign;
  dwWrite=pPlayer->open.ring.dwWritten;
#endif // }

  if (dwWrite<dwSize) {
    if (pPlayer->state<PLAYER_STATE_DRAIN_1) {
      if (bNeedPadding||0==(dwWriteSize=dwWrite)) {
        DPRINTF(debug,0,"  %s: underflow\n",__func__);
        PlayerKill(pPlayer);
        pPlayer->state=PLAYER_STATE_UNDERFLOW;
        goto underflow;
      }
    }

	dwWriteSize=dwWrite;

    if (!bNeedPadding)
      uFramesWrite=dwWriteSize/pwfx->nBlockAlign;
  }
  else
    dwWriteSize=dwSize;

  if (PlayerGetDlgConfig(pPlayer)) {
    wRing=MulDiv(USHRT_MAX,pRing->dwWritten,pRing->pMax-pRing->pRep);
    wShared=MulDiv(USHRT_MAX,uFramesPadding,pPlayer->open.uFrames);
    PlayerPostUpdate(pPlayer,wRing,wShared);
  }

  if (0==uFramesWrite)
    goto nop;

  // get the render buffer ////////////////////////////////////////////////////
  hr=pRender->lpVtbl->GetBuffer(pRender,
    uFramesWrite,     // [in]   UINT32 NumFramesRequested,
    &pData            // [out]  BYTE **ppData
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_BUFFER_ERROR,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_TOO_LARGE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"getting render buffer");
    goto get;
  }

  // copy the data ////////////////////////////////////////////////////////////
#if defined (YASAPI_RING_DWORD) // {
  dwFramesRead=RingReadEx(pRing,pData,dwWriteSize,RING_COPY);

  if (0<dwFramesRead)
    dwFramesRead/=pwfx->nBlockAlign;
#else // } {
  uFramesRead=RingReadEx(pRing,pData,dwWriteSize,RING_COPY);
  uFramesRead/=pwfx->nBlockAlign;
#endif // }

  if (dwWriteSize<dwSize&&bNeedPadding)
    ZeroMemory(pData+dwWriteSize,dwSize-dwWriteSize);

  // release the render buffer ////////////////////////////////////////////////
#if defined (YASAPI_RING_DWORD) // {
  hr=pRender->lpVtbl->ReleaseBuffer(pRender,
    0<dwFramesRead?dwFramesRead:0,
                      // [in]  UINT32 NumFramesWritten,
    dwFlags           // [in]  DWORD dwFlags
  );
#else // } {
  hr=pRender->lpVtbl->ReleaseBuffer(pRender,
    uFramesRead,      // [in]  UINT32 NumFramesWritten,
    dwFlags           // [in]  DWORD dwFlags
  );
#endif // }

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_INVALID_SIZE,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_SIZE_ERROR,hr);
    DERROR(debug,AUDCLNT_E_OUT_OF_ORDER,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"releasing render buffer");
    goto release;
  }

#if defined (YASAPI_RING_DWORD) // {
  if (dwFramesRead<0) {
    DMESSAGE(debug,"reading ring buffer");
    goto read;
  }
#endif // }

  // no critical section needed in order to block unexpected write
  // requests to the ring buffer because all requests are serialized.
  RingReadEx(pRing,pData,dwWriteSize,RING_COMMIT);
nop:
  if (PLAYER_STATE_DRAIN_1==pPlayer->state&&0==pPlayer->open.ring.dwWritten)
    ++pPlayer->state;

  pPlayer->open.pStrategy->SetTimer(pPlayer,pPlayer->open.time,FALSE);
underflow:
state:
  return 0;
#if defined (YASAPI_RING_DWORD) // {
read:
#endif // }
release:
get:
  return -1;
}

static int PlayerTryStart(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  const Strategy *pStrategy=pPlayer->open.pStrategy;
  IAudioClient *pClient=pPlayer->open.pClient;
  WAVEFORMATEX *pwfx=&pPlayer->open.wfxx.Format;
  Ring *pRing=&pPlayer->open.ring;
  UINT32 uFramesRingWritten=pRing->dwWritten/pwfx->nBlockAlign;
  UINT32 uFramesPadding;
  HRESULT hr;

  if (pPlayer->state<PLAYER_STATE_OPEN) {
    DPRINTF(debug,0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }

  if ((uFramesPadding=pStrategy->GetFramesPadding(pPlayer))<0) {
    DMESSAGE(debug,"getting frames padding");
    goto padding;
  }

  if (uFramesRingWritten+uFramesPadding<pPlayer->open.uFramesMin) {
    DPRINTF(debug,0,"  %s: underflow\n",__func__);
    goto underflow;
  }

  // start the audio client ///////////////////////////////////////////////////
  hr=pClient->lpVtbl->Start(pClient);

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_NOT_STOPPED,hr);
    DERROR(debug,AUDCLNT_E_EVENTHANDLE_NOT_SET,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"starting audio client");
    goto start;
  }

  DPRINTF(debug,0,"  %s: audio client started\n",__func__);

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->state=PLAYER_STATE_PLAY;

  /////////////////////////////////////////////////////////////////////////////
  if (PlayerPlay(pPlayer,uFramesPadding)<0) {
    DMESSAGEV(debug,"playing");
    goto play;
  }

  DPRINTF(debug,0,"  %s: start playing\n",__func__);
underflow:
  return 0;
play:
start:
padding:
  pPlayer->state=PLAYER_STATE_OPEN;
state:
  return -1;
}

int PlayerWrite(Player *pPlayer, Request *pRequest)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  const char *buf=va_arg(pRequest->ap,const char *);
  int len=va_arg(pRequest->ap,int);
  Ring *pRing=&pPlayer->open.ring;

  if (pPlayer->state<PLAYER_STATE_OPEN) {
    DPRINTF(debug,0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }

#if defined (YASAPI_RING_REALLOC) // {
  if (RingReallocAvailable(pRing,len,pPlayer->open.nShift)<0) {
    DMESSAGE(debug,"reallocating");
    goto realloc;
  }
#endif // }

#if defined (YASAPI_RING_DWORD) // {
  if (RingWrite(pRing,buf,len,pPlayer->open.nShift)<0) {
    DMESSAGE(debug,"writing");
    goto write;
  }
#else // } {
  RingWrite(pRing,buf,len,pPlayer->open.nShift);
#endif // }

  if (pPlayer->state<PLAYER_STATE_PLAY)
    PlayerTryStart(pPlayer);

  return 0;
#if defined (YASAPI_RING_DWORD) // {
write:
#endif // }
#if defined (YASAPI_RING_REALLOC) // {
realloc:
#endif // }
//ring:
state:
  return -1;
}

int PlayerFlush(Player *pPlayer, Request *pRequest)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  IAudioClient *pClient=pPlayer->open.pClient;
  HRESULT hr;

  if (pPlayer->state<PLAYER_STATE_OPEN) {
    DPRINTF(debug,0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }

  PlayerKill(pPlayer);
  RingReset(&pPlayer->open.ring);
  
  hr=pClient->lpVtbl->Reset(pClient);

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_NOT_STOPPED,hr);
    DERROR(debug,AUDCLNT_E_BUFFER_OPERATION_PENDING,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"re-setting audio client");
    goto reset;
  }

#if defined (YASAPI_GAPLESS) // {
  if (PlayerResetGaplessOffset(pPlayer)<0) {
    DMESSAGE(debug,"re-setting gapless offset\n");
    goto offset;
  }
#endif // }

  return PlayerGetTime(pPlayer,NULL);
#if defined (YASAPI_GAPLESS) // {
offset:
#endif // }
reset:
state:
  return -1;
}

int PlayerIsPlaying(Player *pPlayer, Request *pRequest)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }

  if (pPlayer->state<PLAYER_STATE_OPEN) {
    DPRINTF(debug,0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    return 0;
  }
  else if (pPlayer->open.ring.dwWritten) {
    pPlayer->state=PLAYER_STATE_DRAIN_1;
    DPUTS(debug,0,"  isplaying\n");
    return 1;
  }
  else {
    switch (pPlayer->state) {
    case PLAYER_STATE_PLAY:
      pPlayer->state=PLAYER_STATE_DRAIN_2;
      DPUTS(debug,0,"  isplaying\n");
      return 1;
    case PLAYER_STATE_DRAIN_1:
    case PLAYER_STATE_DRAIN_2:
    case PLAYER_STATE_DRAIN_3:
      DPRINTF(debug,0,"  isplaying (drain %d)\n",
          1+pPlayer->state-PLAYER_STATE_DRAIN_1);
      return 1;
    case PLAYER_STATE_EOT:
    default:
      DPUTS(debug,0,"  isplaying (eot)\n");
      return 0;
    }
  }
}

int PlayerPause(Player *pPlayer, Request *pRequest)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  int pause=va_arg(pRequest->ap,int);
  int time;

  if (pPlayer->state<PLAYER_STATE_OPEN) {
    DPRINTF(debug,0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }
  else if (pause) {
    PlayerKill(pPlayer);

#if defined (YASAPI_GAPLESS) // {
    if (PlayerResetGaplessOffset(pPlayer)<0) {
      DMESSAGE(debug,"re-setting gapless offset\n");
      goto offset;
    }
#endif // }

    pPlayer->state=PLAYER_STATE_PAUSE;
  }
  else {
    if (PlayerTryStart(pPlayer)<0) {
      DMESSAGE(debug,"trying to start the player");
      goto start;
    }
  }

  return PlayerGetTime(pPlayer,NULL);
start:
  return 0;
#if defined (YASAPI_GAPLESS) // {
offset:
#endif // }
state:
  return -1;
}

int PlayerGetTime(Player *pPlayer, Request *pRequest)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
#if defined (YASAPI_GAPLESS) // {
  WAVEFORMATEXTENSIBLE *pwfxx=&pPlayer->open.wfxx;
  WAVEFORMATEX *pwfx=&pwfxx->Format;
  int nChannels=pwfx->nChannels;
  int nSamplesPerSec=pwfx->nSamplesPerSec;
  int bPosition
      =YASAPI_GAPLESS_OFFSET_POSITION==pPlayer->options.device.nGaplessOffset;
#endif // }
  UINT64 u64Position;
  UINT64 u64Frequency;
  double qTime;

  if (pPlayer->state<PLAYER_STATE_OPEN) {
    DPRINTF(debug,0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }

  if ((u64Position=PlayerGetPosition(pPlayer))<0) {
    DMESSAGE(debug,"getting position");
    goto position;
  }

#if defined (YASAPI_GAPLESS)
  if (bPosition)
    u64Frequency=pPlayer->open.u64Frequency;
  else {
#endif
    if ((u64Frequency=PlayerGetFrequency(pPlayer))<=0) {
      DMESSAGE(debug,"getting frequency");
      goto frequency;
    }
#if defined (YASAPI_GAPLESS)
  }
#endif

  qTime=1000.0*u64Position/u64Frequency;

#if defined (YASAPI_GAPLESS) // {
  if (bPosition)
    qTime-=1000.0*pPlayer->open.u64Position/u64Frequency;
  else
    qTime-=1000.0*pPlayer->open.qTime;
#endif // }

  return qTime+0.5;
frequency:
position:
state:
  return 0;
}

#if defined (YASAPI_GAPLESS) // {
int PlayerSetState(Player *pPlayer, Request *pRequest)
{
  int state=va_arg(pRequest->ap,int);

  if (PLAYER_STATE_RESET==state)
    pPlayer->state_save=pPlayer->state;

  return pPlayer->state=state;
}
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_RING_REALLOC) // {
static PlayerWriteArgs(Player *pPlayer, Result *pResult, ...)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  Request request;
  va_list ap;
  int state;

  ZeroMemory(&request, sizeof request);
  va_start(ap,pResult);
  request.pResult=pResult;
  request.ap=ap;
  pResult->state=PlayerWrite(pPlayer,&request);
  va_end(ap);
}
#endif // }

int PlayerRead(Player *pPlayer, Request *pRequest)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  UINT32 uFramesPadding;

  if (pPlayer->state<PLAYER_STATE_OPEN) {
    DPRINTF(debug,0,"  Warning: unexpected state: %d (%s, \"%s\", %d)\n",
        pPlayer->state,__func__,basename(__FILE__),__LINE__);
    goto state;
  }
  else if (PLAYER_STATE_DRAIN_2==pPlayer->state) {
    pPlayer->state=PLAYER_STATE_DRAIN_3;
    pPlayer->open.pStrategy->SetTimer(pPlayer,pPlayer->open.time,TRUE);
  }
  else if (PLAYER_STATE_DRAIN_3==pPlayer->state) {
    pPlayer->state=PLAYER_STATE_EOT;
    goto eof;
  }
  else if (PLAYER_STATE_PAUSE==pPlayer->state)
    goto pause;
  else {
    if ((uFramesPadding=PlayerGetFramesPadding(pPlayer))<0) {
      DMESSAGE(debug,"getting frames padding");
      goto padding;
    }

    if (PlayerPlay(pPlayer,uFramesPadding)<0) {
      DMESSAGE(debug,"playing");
      goto play;
    }
  }
pause:
eof:
  return 0;
play:
padding:
state:
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
VOID CALLBACK PlayerFireTimer(LPVOID lpArg, DWORD dwLow, DWORD dwHigh)
{
  PlayerDispatchAsync(lpArg);
}

void PlayerSetTimer(Player *pPlayer, LONGLONG time)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  HANDLE hTimer=pPlayer->base.hTimer;
  LARGE_INTEGER ui;

  // http://msdn.microsoft.com/en-us/library/windows/desktop/ms687012%28v=vs.85%29.aspx
  // Positive values indicate absolute time. 
  // Negative values indicate relative time.
  // in 100 nanosecond intervals
  ui.QuadPart=-time;

  SetWaitableTimer(
    hTimer,           // _In_      HANDLE hTimer,
    &ui,              // _In_      const LARGE_INTEGER *pDueTime,
    0,                // _In_      LONG lPeriod,
    time?PlayerFireTimer:NULL,
                      // _In_opt_  PTIMERAPCROUTINE pfnCompletionRoutine,
    time?pPlayer:NULL,
                      // _In_opt_  LPVOID lpArgToCompletionRoutine,
    FALSE             // _In_      BOOL fResume
  );
}

UINT32 PlayerGetFramesPadding(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  IAudioClient *pClient=pPlayer->open.pClient;
  UINT32 uFramesPadding;
  HRESULT hr;

  hr=pClient->lpVtbl->GetCurrentPadding(pClient,
    &uFramesPadding       // [out]  UINT32 *pNumPaddingFrames
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DUNKNOWN(debug,hr);
    DMESSAGE(debug,"getting padding");
    goto padding;
  }

  return uFramesPadding;
padding:
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
HWND PlayerGetDlgConfig(Player *pPlayer)
{
  return pPlayer->options.common.bVisualization?pPlayer->hDlgConfig:NULL;
}

void PlayerPostUpdate(Player *pPlayer, WORD wRing, WORD wShared)
{
  HWND hDlgConfig=PlayerGetDlgConfig(pPlayer);

  if (hDlgConfig)
    PostMessage(hDlgConfig,WM_CONFIG_UPDATE,0,MAKELONG(wRing,wShared));
}

///////////////////////////////////////////////////////////////////////////////
int PlayerDispatchSync(Player *pPlayer, int exit, player_proc_t PlayerProc,
    ...)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  Store *pStore=&pPlayer->base.store;
  Queue *pQueue=&pPlayer->base.queue;
  Result *pResult;
  Request *pRequest;
  int state;
  va_list ap;

  va_start(ap,PlayerProc);
  pResult=StoreGet(pStore);
  pRequest=QueueLockWrite(pQueue,pResult,exit,PlayerProc,pPlayer,ap);
  QueueUnlockWrite(pQueue,pResult);
    state=pResult->state;
  StorePut(pStore,pResult);
  va_end(ap);

  return state;
}

void PlayerDispatchAsync(void *data)
{
  Player *pPlayer=data;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  Queue *pQueue=&pPlayer->base.queue;
  Request *pRequest;

  QueueLockWrite(pQueue,NULL,0,PlayerRead,pPlayer,NULL);
  QueueUnlockWrite(pQueue,NULL);
}

///////////////////////////////////////////////////////////////////////////////
static HANDLE PlayerGetEvent(Player *pPlayer)
{
  return pPlayer->state<PLAYER_STATE_PLAY
      ?NULL:pPlayer->open.pStrategy->GetEvent(pPlayer);
}

DWORD WINAPI PlayerThread(LPVOID lpParameter)
{
  Player *pPlayer=lpParameter;
  Queue *pQueue=&pPlayer->base.queue;
  int exit=0;
  DWORD dwTaskIndex=0;
  HANDLE hTask=AvSetMmThreadCharacteristicsW(L"Pro Audio",&dwTaskIndex);
  HANDLE hEvent;
  Request *pRequest;
  int state;
#if defined (YASAPI_THREAD_CO_INITIALIZE) // {
#if defined (YASAPI_DEBUG) // {
  int nDebug=pPlayer->options.common.nDebug;
#endif // }
  HRESULT hr;

  hr=CoInitializeEx(
    NULL,                     // _In_opt_ LPVOID pvReserved
	  COINIT_MULTITHREADED		  // _In_     DWORD  dwCoInit
  );

  if (FAILED(hr)) {
    DERROR(nDebug,S_FALSE,hr);
    DERROR(nDebug,RPC_E_CHANGED_MODE,hr);
    DMESSAGE(nDebug,"initializing com failed");
    goto coinit;
  }
#endif // }

  while (!exit) {
    hEvent=PlayerGetEvent(pPlayer);
    pRequest=QueueLockRead(pQueue,hEvent,PlayerDispatchAsync,pPlayer);
    state=pRequest->PlayerProc(pPlayer,pRequest);
    exit=pRequest->exit;

    if (NULL!=pRequest->pResult)
      pRequest->pResult->state=state;

    QueueUnlockRead(&pPlayer->base.queue,pRequest->pResult);
  }

#if defined (YASAPI_THREAD_CO_INITIALIZE) // {
  CoUninitialize();
coinit:
#endif // }

  if (hTask)
    AvRevertMmThreadCharacteristics(hTask);

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
void PlayerCopyMemory(PVOID p, PVOID Destination, const VOID *Source,
    SIZE_T Length)
{
  Player *pPlayer=p;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  Options *pOptions=&pPlayer->options;
  double qVolume=pPlayer->base.qVolume;
  char *wp,*mp;
  const char *rp;
  int m,n;
  double q;
  int32_t i32;

  wp=Destination;
  mp=wp+Length;
  rp=Source;
  // bits per sample.
  n=pPlayer->open.wfxx.Format.wBitsPerSample;
  // bytes per sample.
  m=n>>3;
  // attenuation factor.
  q=qVolume*YASAPI_INT_MAX(n)/YASAPI_INT_MAX(32);
  // bits to shift the 32-bit result.
  n=32-n;

  if (!pPlayer->open.nShift) {
    if (!pOptions->common.bVolume||1.0==qVolume) {
//fprintf(stderr,"  %s 1\n",__func__);
      CopyMemory(
        Destination,            // _In_  PVOID Destination,
        Source,                 // _In_  const VOID *Source,
        Length                  // _In_  SIZE_T Length
      );
//fprintf(stderr,"  %s 2\n",__func__);
    }
    else {
//fprintf(stderr,"  %s 3\n",__func__);
      while (wp<mp) {
        i32=0;

        CopyMemory(
          &i32,                 // _In_  PVOID Destination,
          rp,                   // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        i32<<=n;
        i32=q*i32+0.5;

        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          &i32,                 // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;
        rp+=m;
      }
//fprintf(stderr,"  %s 4\n",__func__);
    }
  }
  else {
    if (!pOptions->common.bVolume||1.0==qVolume) {
//fprintf(stderr,"  %s 5\n",__func__);
      while (wp<mp) {
        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          rp,                   // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;

        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          rp,                   // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;
        rp+=m;
      }
//fprintf(stderr,"  %s 6\n",__func__);
    }
    else {
//fprintf(stderr,"  %s 7\n",__func__);
      while (wp<mp) {
        i32=0;

        CopyMemory(
          &i32,                 // _In_  PVOID Destination,
          rp,                   // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        i32<<=n;
        i32=q*i32+0.5;

        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          &i32,                 // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;

        CopyMemory(
          wp,                   // _In_  PVOID Destination,
          &i32,                 // _In_  const VOID *Source,
          m                     // _In_  SIZE_T Length
        );

        wp+=m;
        rp+=m;
      }
//fprintf(stderr,"  %s 8\n",__func__);
    }
  }
}
